
#include "stdio.h"
#include "stdlib.h"
#include "math.h"
#include <string.h>

#include "Algorithme1.h"

/* --------------------------------------------------------------------------- *
 * Ce programme est l'implémentation de l'algorithme du chemin aléatoire. La
 * fonction communitiesPartition prend comme argument un tableau d'entiers à
 * deux dimensions qui correspond à la matrice d'adjacence du graphe, elle prend
 * également comme argument un entier n qui correspond à la taille de la matrice
 * d'adjacence et un entier t qui est la paramètre de l'algorithme. Lorsque
 * l'algorithme est terminé, la fonction écrit dans un fichier nommé
 * "meilleureCom.txt" les différentes communautés détectées dans le graphe ainsi
 * que la valeur maximale de eta_k.
 * -------------------------------------------------------------------------- */



static EnsembleDesCommunautes creationCommunautes(int n) {
    int i;

    EnsembleDesCommunautes newEnsembleCom;
    newEnsembleCom.nombreCommunautes = n;

    Communaute *vecteurCom;
    vecteurCom = (Communaute*)malloc(n*sizeof(Communaute));
	if(vecteurCom==NULL){
		printf("Allocation impossible\n");
		exit(1);
	}

    for(i=0; i<n; i++){
        Communaute Com;
        Com.numero = i;
        Com.nombreElements = 1;
        int* sommets = (int*)malloc(1*sizeof(int));
        sommets[0] = i;
        Com.listeSommets = sommets;
        vecteurCom[i]= Com;
    }
    newEnsembleCom.vecteurCommunautes = vecteurCom;
    return newEnsembleCom;
}

static Communaute copieCommunaute(Communaute com){
    size_t i;
    Communaute newCommunaute;
    newCommunaute.numero = com.numero;
    newCommunaute.nombreElements = com.nombreElements;
    int * sommets = (int*)malloc((newCommunaute.nombreElements)*sizeof(int));
    for(i=0; i<newCommunaute.nombreElements; i++){
        sommets[i] = com.listeSommets[i];
    }
    newCommunaute.listeSommets = sommets;
    return newCommunaute;
}

static EnsembleDesCommunautes copieEnsembleCommunaute(EnsembleDesCommunautes ensembleInitialeCommunautes){
    size_t i;
    EnsembleDesCommunautes newEnsembleCom;
    size_t n = ensembleInitialeCommunautes.nombreCommunautes;
    newEnsembleCom.nombreCommunautes = n;

    Communaute *vecteurCom;
    vecteurCom = (Communaute*)malloc(n*sizeof(Communaute));
    if(vecteurCom==NULL){
        printf("Allocation impossible\n");
        exit(1);
    }

    for(i=0;i<n;i++){
        Communaute Com = copieCommunaute(ensembleInitialeCommunautes.vecteurCommunautes[i]);
        vecteurCom[i] = Com;
    }
    newEnsembleCom.vecteurCommunautes = vecteurCom;
    return newEnsembleCom;
}

static void ecritureEnsComDansFichier(EnsembleDesCommunautes meilleureCom, char* nomDeFichier, double eta){
    size_t i,j;
    FILE* fichierEcriture = NULL;
    fichierEcriture = fopen(nomDeFichier,"w");
    if(fichierEcriture==NULL){
        printf("Impossible d'ouvrir le fichier \n");
        exit(1);
    }

    char text1[128] = "eta = ";
    char text2[128] = "";

    sprintf(text2,"%1.20f",eta);
    strcat(text1,text2);
    strcat(text1,"\n");
    fputs(text1,fichierEcriture);

    fprintf(fichierEcriture, "nombre de communautes = %zu \n", meilleureCom.nombreCommunautes);
    for(i=0; i<meilleureCom.nombreCommunautes; i++){
        fprintf(fichierEcriture, "com %zu = numero - %zu, nombreElements - %zu, listeSommets - ", i, meilleureCom.vecteurCommunautes[i].numero, meilleureCom.vecteurCommunautes[i].nombreElements);
        for(j=0; j<meilleureCom.vecteurCommunautes[i].nombreElements; j++){
            fprintf(fichierEcriture, "%d - ", meilleureCom.vecteurCommunautes[i].listeSommets[j]);
        }
        fprintf(fichierEcriture, "\n");
    }

    fclose(fichierEcriture);
}

static void freeMatriceDouble(double** matrice, int n){
    int i;
    for(i=0; i<n; i++){
        free(matrice[i]);
    }
    free(matrice);
}

static int degre(int** matriceAdjacence, int n, int numeroLigne){
    int i;
    int somme = 0;
    for (i=0; i<n; i++){
        somme = somme + matriceAdjacence[numeroLigne][i];
    }
    return somme;
}

static int* creationVecteurDegres(int** matriceAdjacence, int n){
    int i;
    int* vecteurDegres = (int*)malloc(n*sizeof(int));
	if(vecteurDegres==NULL){   //test si l'allocation a reussi
		printf("Allocation impossible\n");
		exit(1);
	}

    for(i=0; i<n; i++){
        vecteurDegres[i] = degre(matriceAdjacence, n, i);
    }
    return vecteurDegres;
}

static double** creationMatriceTransition(int** matriceAdjacence, int* vecteurDegres, int n){
    int i,j;
    double ** matriceTransition = (double**)malloc(n*sizeof(double*));
	if(matriceTransition==NULL){
		printf("Allocation impossible\n");
		exit(1);
	}

	for(i=0; i<n; i++){
		matriceTransition[i] = (double*)malloc(n*sizeof(double));
		if(matriceTransition[i]==NULL){
			printf("Allocation impossible\n");
			exit(1);
		}
	}

    for(i=0; i<n; i++){
        double d = vecteurDegres[i];
        for (j=0; j<n; j++){
            matriceTransition[i][j] = matriceAdjacence[i][j]/d ;
        }
    }
    return matriceTransition;
}

static double** produitDeDeuxMatrices(double** A, double** B, int n){
    int i,j,k;
    double ** C = (double**)malloc(n*sizeof(double*));
	if(C==NULL){
		printf("Allocation impossible\n");
		exit(1);
	}

	for(i=0; i<n; i++){
		C[i] = (double*)malloc(n*sizeof(double));
		if(C[i]==NULL){
			printf("Allocation impossible\n");
			exit(1);
		}
	}
    for(i=0; i<n; i++){
        for(j=0; j<n; j++){
            double s = 0;
            for(k=0; k<n; k++){
                s = s + A[i][k]*B[k][j];
            }
            C[i][j]=s;
        }
    }
    return C;
}

static double** matriceTransitionExposant(double** matriceTransition, int n, int t){
    int i;
    double** exposantMatrice = (double**)malloc(n*sizeof(double*));
	if(exposantMatrice==NULL){
		printf("Allocation impossible\n");
		exit(1);
	}

	for(i=0; i<n; i++){
		exposantMatrice[i]=(double*)malloc(n*sizeof(double));
		if(exposantMatrice[i]==NULL){
			printf("Allocation impossible\n");
			exit(1);
		}
	}

    exposantMatrice = produitDeDeuxMatrices(matriceTransition, matriceTransition, n);

    for(i=2; i<t; i++){
        exposantMatrice = produitDeDeuxMatrices(exposantMatrice, matriceTransition, n);
    }
    return exposantMatrice;
}

static double** creationMatriceDistancesSommets(int** matriceAdjacence, double** exposantMatriceTransition, int* vecteurDegres, int n){
    int i,j,k;
    double ** matriceDistance = (double**)malloc(n*sizeof(double*));
	if(matriceDistance==NULL){
		printf("Allocation impossible\n");
		exit(1);
	}

	for(i=0; i<n; i++){
		matriceDistance[i] = (double*)malloc(n*sizeof(double));
		if(matriceDistance[i]==NULL){
			printf("Allocation impossible\n");
			exit(1);
		}
	}

    for(i=0; i<n; i++){
        for(j=0; j<n; j++){
            double somme = 0;
            if(matriceAdjacence[i][j] != 0){
                for(k=0; k<n; k++){
                    somme = somme + (pow((exposantMatriceTransition[i][k]-exposantMatriceTransition[j][k]),2)/vecteurDegres[k]);
                }
                matriceDistance[i][j] = sqrt(somme);
            }
            else{
                matriceDistance[i][j] = INFINITY;
            }
        }
    }
    return matriceDistance;
}

static double distanceEntreDeuxCommunautes(Communaute com1, Communaute com2, int* vecteurDegres, double** exposantMatriceTransition, int n){
    int k;
    double somme = 0;
    for(k=0; k<n; k++){
        double pc1 = exposantMatriceTransition[com1.numero][k];
        double pc2 = exposantMatriceTransition[com2.numero][k];
        somme = somme + (pow((pc1-pc2),2)/vecteurDegres[k]);
    }
    return sqrt(somme);
}

static double variationSigma(Communaute com1, Communaute com2, int n, int* vecteurDegres, double** exposantMatriceTransition){
    double r;
    r = distanceEntreDeuxCommunautes(com1, com2, vecteurDegres, exposantMatriceTransition, n);
    double deltaSigma = r*r*(com1.nombreElements*com2.nombreElements)/(n*(com1.nombreElements+com2.nombreElements));
    return deltaSigma;
}

static int communauteAdjacente(int** matriceAdjacence, Communaute com1, Communaute com2){
    size_t i,j;
    int adjacent = 0;
    for(i=0; i<com1.nombreElements; i++){
        for(j=0; j<com2.nombreElements; j++){
            if(adjacent==0){
                if(matriceAdjacence[com1.listeSommets[i]][com2.listeSommets[j]] != 0){
                    adjacent = 1;
                }
            }
        }
    }
    return adjacent;
}

static double** creationMatriceDeltaSigma(int** matriceAdjacence, double** matriceDistanceSommets, int n){
    int i,j;
    double ** matriceDeltaSigma = (double**)malloc(n*sizeof(double*));
	if(matriceDeltaSigma==NULL){
		printf("Allocation impossible\n");
		exit(1);
	}

	for(i=0; i<n; i++){
		matriceDeltaSigma[i] = (double*)malloc(n*sizeof(double));
		if(matriceDeltaSigma[i]==NULL){
			printf("Allocation impossible\n");
			exit(1);
		}
	}

    for(i=0; i<n ; i++){
        for(j=0; j<n; j++){
            if(matriceAdjacence[i][j]!=0){
                double r = matriceDistanceSommets[i][j];
                matriceDeltaSigma[i][j] = (r*r)/(2*n);
            }
            else{
                matriceDeltaSigma[i][j] = INFINITY;
            }
        }
    }
    return matriceDeltaSigma;
}

static double* minimumSaufDiagonale(double** matriceDeltaSigma, int n){
    int i,j;
    double* ligneColonneMin = (double*)malloc(3*sizeof(double));
    if(ligneColonneMin==NULL){
		printf("Allocation impossible\n");
		exit(1);
	}

    ligneColonneMin[0] = 0;
    ligneColonneMin[1] = 1;
    double min = matriceDeltaSigma[0][1];
    ligneColonneMin[2] = min;

    for(i=0; i<n; i++){
        for(j=0; j<n; j++){
            if(i!=j){
                if(matriceDeltaSigma[i][j]<min){
                    min = matriceDeltaSigma[i][j];
                    ligneColonneMin[0] = i;
                    ligneColonneMin[1] = j;
                    ligneColonneMin[2] = min;
                }
            }
        }
    }
    return ligneColonneMin;
}

static double LWJformula(Communaute com1, Communaute com2, Communaute com3, double** matriceDeltaSigma, int n, int* vecteurDegres, double** exposantMatriceTransition){
    int nbr1 = com1.nombreElements;
    int nbr2 = com2.nombreElements;
    int nbr3 = com3.nombreElements;
    int num1 = com1.numero;
    int num2 = com2.numero;
    int num3 = com3.numero;

    double delta1, delta2, delta3;

    if(matriceDeltaSigma[num1][num3] == INFINITY){
        double newDeltaSigma = variationSigma(com1, com3, n, vecteurDegres, exposantMatriceTransition);
        matriceDeltaSigma[num1][num3] = newDeltaSigma;
        matriceDeltaSigma[num3][num1] = newDeltaSigma;
        delta1 = newDeltaSigma;
    }
    else{
        delta1 = matriceDeltaSigma[num1][num3];
    }

    if(matriceDeltaSigma[num2][num3] == INFINITY){
        double newDeltaSigma = variationSigma(com2, com3, n, vecteurDegres, exposantMatriceTransition);
        matriceDeltaSigma[num2][num3] = newDeltaSigma;
        matriceDeltaSigma[num3][num2] = newDeltaSigma;
        delta2 = newDeltaSigma;
    }
    else{
        delta2 = matriceDeltaSigma[num2][num3];
    }

    if(matriceDeltaSigma[num1][num2] == INFINITY){
        double newDeltaSigma = variationSigma(com1, com2, n, vecteurDegres, exposantMatriceTransition);
        matriceDeltaSigma[num1][num2] = newDeltaSigma;
        matriceDeltaSigma[num2][num1] = newDeltaSigma;
        delta3 = newDeltaSigma;
    }
    else{
        delta3 = matriceDeltaSigma[num1][num2];
    }

    double n1 = (nbr1 + nbr3)*delta1;
    double n2 = (nbr2 + nbr3)*delta2;
    double n3 = nbr3*delta3;

    return (n1+n2-n3)/(nbr1+nbr2+nbr3);
}

static double** modificationMatriceTransition(double** exposantMatriceTransition, int n, int newTaille, Communaute com1, Communaute com2){
    int i,j;
    double ** newExposantMatriceTransition = (double**)malloc(newTaille*sizeof(double*));
    if(newExposantMatriceTransition==NULL){
        printf("Allocation impossible\n");
        exit(1);
    }

    for(i=0; i<newTaille; i++){
        newExposantMatriceTransition[i] = (double*)malloc(n*sizeof(double));
        if(newExposantMatriceTransition[i]==NULL){
            printf("Allocation impossible\n");
            exit(1);
        }
    }

    int petit;
    int grand;
    if(com1.numero<com2.numero){
        petit = com1.numero;
        grand = com2.numero;
    }
    else{
        petit = com2.numero;
        grand = com1.numero;
    }

    double* vecteurP = (double*)malloc(n*sizeof(double));
    if(vecteurP==NULL){
        printf("Allocation impossible\n");
        exit(1);
    }

    int nbr1 = com1.nombreElements;
    int nbr2 = com2.nombreElements;

    for (j=0; j<n; j++){
        vecteurP[j] = ((nbr1*exposantMatriceTransition[com1.numero][j])+(nbr2*exposantMatriceTransition[com2.numero][j]))/(nbr1+nbr2);
    }

    for(i=0; i<newTaille; i++){
        for (j=0; j<n; j++){
            if(i==petit){
                newExposantMatriceTransition[i][j] = vecteurP[j];
            }
            else if(i<grand){
                    newExposantMatriceTransition[i][j] = exposantMatriceTransition[i][j];
            }
            else{
                    newExposantMatriceTransition[i][j] = exposantMatriceTransition[i+1][j];
            }
        }
    }
    freeMatriceDouble(exposantMatriceTransition, newTaille+1);
    return newExposantMatriceTransition;
}

static double** modificationDeltaSigma(int** matriceAdjacence, EnsembleDesCommunautes ensembleCommunautes, double** matriceDeltaSigma, int newTaille, Communaute com1, Communaute com2, int n, int* vecteurDegres, double** exposantMatriceTransition){
    int i=0;
    int j=0;
    double ** newMatriceDeltaSigma = (double**)malloc(newTaille*sizeof(double*));
	if(newMatriceDeltaSigma==NULL){
		printf("Allocation impossible\n");
		exit(1);
	}

	for(i=0; i<newTaille; i++){
		newMatriceDeltaSigma[i] = (double*)malloc(newTaille*sizeof(double));
		if(newMatriceDeltaSigma[i]==NULL){
			printf("Allocation impossible\n");
			exit(1);
		}
	}

    double* vecteurDelta = (double*)malloc(newTaille*sizeof(double));
    if(vecteurDelta==NULL){
        printf("Allocation impossible\n");
        exit(1);
    }

    int petit, grand;
    if(com1.numero<com2.numero){
        petit = com1.numero;
        grand = com2.numero;
    }
    else{
        petit = com2.numero;
        grand = com1.numero;
    }

    for (i=0; i<newTaille; i++){
        if(i==petit){
            vecteurDelta[i] = 0;
        }
        else{
            if(i<grand){
                int adjacent1 = communauteAdjacente(matriceAdjacence, com1, ensembleCommunautes.vecteurCommunautes[i]);
                int adjacent2 = communauteAdjacente(matriceAdjacence, com2, ensembleCommunautes.vecteurCommunautes[i]);
                if(adjacent1 == 0 && adjacent2 == 0){
                    vecteurDelta[i] = INFINITY;
                }
                else{
                    vecteurDelta[i] = LWJformula(com1, com2, ensembleCommunautes.vecteurCommunautes[i], matriceDeltaSigma, n, vecteurDegres, exposantMatriceTransition);
                }
            }
            else{
                int adjacent1 = communauteAdjacente(matriceAdjacence, com1, ensembleCommunautes.vecteurCommunautes[i+1]);
                int adjacent2 = communauteAdjacente(matriceAdjacence, com2, ensembleCommunautes.vecteurCommunautes[i+1]);
                if(adjacent1 == 0 && adjacent2 == 0){
                    vecteurDelta[i] = INFINITY;
                }
                else{
                    vecteurDelta[i] = LWJformula(com1, com2, ensembleCommunautes.vecteurCommunautes[i+1], matriceDeltaSigma, n, vecteurDegres, exposantMatriceTransition);
                }
            }
        }
    }

    for(i=0; i<newTaille; i++){
        for (j=0; j<newTaille; j++){
            if(i==petit){
                newMatriceDeltaSigma[i][j] = vecteurDelta[j];
            }
            else if(i<grand){
                if(j==petit){
                    newMatriceDeltaSigma[i][j] = vecteurDelta[i];
                }
                else if(j<grand){
                    newMatriceDeltaSigma[i][j] = matriceDeltaSigma[i][j];
                }
                else{
                    newMatriceDeltaSigma[i][j] = matriceDeltaSigma[i][j+1];
                }
            }
            else{
                if(j==petit){
                    newMatriceDeltaSigma[i][j] = vecteurDelta[i];
                }
                else if(j<grand){
                    newMatriceDeltaSigma[i][j] = matriceDeltaSigma[i+1][j];
                }
                else{
                    newMatriceDeltaSigma[i][j] = matriceDeltaSigma[i+1][j+1];
                }
            }
        }
    }
    freeMatriceDouble(matriceDeltaSigma, newTaille+1);
    return newMatriceDeltaSigma;
}

static EnsembleDesCommunautes fusionCommunautes(EnsembleDesCommunautes ensembleCommunautes, Communaute com1, Communaute com2){
    size_t i;
    Communaute newCommunaute;
    int nbr = com1.nombreElements + com2.nombreElements;
    newCommunaute.nombreElements = nbr;

    int* sommets = (int*)malloc(nbr*sizeof(int));
    for(i=0; i<com1.nombreElements; i++){
        sommets[i] = com1.listeSommets[i];
    }
    for(i=com1.nombreElements; i<com1.nombreElements+com2.nombreElements; i++){
        sommets[i] = com2.listeSommets[i-com1.nombreElements];
    }
    newCommunaute.listeSommets = sommets;

    int numPetit;
    int numGrand;
    if(com1.numero<com2.numero){
        numPetit = com1.numero;
        numGrand = com2.numero;
    }
    else{
        numPetit = com2.numero;
        numGrand = com1.numero;
    }
    newCommunaute.numero = numPetit;
    ensembleCommunautes.vecteurCommunautes[numPetit] = newCommunaute;
    for(i=numGrand; i<ensembleCommunautes.nombreCommunautes-1; i++){
        ensembleCommunautes.vecteurCommunautes[i] = ensembleCommunautes.vecteurCommunautes[i+1];
        ensembleCommunautes.vecteurCommunautes[i].numero = ensembleCommunautes.vecteurCommunautes[i].numero - 1;
    }
    ensembleCommunautes.nombreCommunautes = ensembleCommunautes.nombreCommunautes - 1;
    return ensembleCommunautes;
}

void communitiesPartition(int** matriceAdjacence, int n, int t){
    int i;
    double eta = 0;
    double sigmakMoinsUn;
    double sigmak;

    EnsembleDesCommunautes com = creationCommunautes(n);
    EnsembleDesCommunautes meilleureCom;
    int* D = creationVecteurDegres(matriceAdjacence, n);
    double** Pinitial = creationMatriceTransition(matriceAdjacence, D, n);
    double** P = matriceTransitionExposant(Pinitial, n, t);
    double** R = creationMatriceDistancesSommets(matriceAdjacence, P, D, n);
    double** S = creationMatriceDeltaSigma(matriceAdjacence, R, n);
    freeMatriceDouble(R, n);
    double* min = (double*)malloc(3*sizeof(double));
    if(min==NULL){
		printf("Allocation impossible\n");
		exit(1);
	}

    for(i=0; i<n-1; i++){

        min = minimumSaufDiagonale(S, n-i);

        if(i==0){
            sigmakMoinsUn = min[2];
        }
        else if(i==1){
            sigmak = min[2];
            if(sigmakMoinsUn>0){
                eta = sigmak/sigmakMoinsUn;
                meilleureCom = copieEnsembleCommunaute(com);
            }
        }
        else {
            sigmakMoinsUn = sigmak;
            sigmak = min[2];
            if(sigmakMoinsUn>0){
                double newEta = sigmak/sigmakMoinsUn;
                if(newEta>=eta){
                    meilleureCom = copieEnsembleCommunaute(com);
                    eta = newEta;
                }
            }
        }

        int numCom1 = min[0];
        int numCom2 = min[1];

        S = modificationDeltaSigma(matriceAdjacence, com, S, n-i-1, com.vecteurCommunautes[numCom1], com.vecteurCommunautes[numCom2], n, D, P);

        P = modificationMatriceTransition(P, n, n-i-1, com.vecteurCommunautes[numCom1], com.vecteurCommunautes[numCom2]);

        com = fusionCommunautes(com, com.vecteurCommunautes[numCom1], com.vecteurCommunautes[numCom2]);
    }

    ecritureEnsComDansFichier(meilleureCom, "meilleureCom.txt", eta);

    free(D);
    freeMatriceDouble(P,1);
    freeMatriceDouble(S,1);
}
